//====================--- I N N E R B R E E D F X ---=========================
//Project; Animatronic Owl for InnerbreedFX
//Description: Control code for Jonny Poole - InnerbreedFX Animatronics
//Software version: V1.0
//Date: 18-09-2012
//Programmer/s: Eric Nantel, James Clifford
//Project Developer/s: Jonny Poole, Eric Nantel
//
//
// This version of the code runs on the DiaLFonZo-Copter Shield with ProMini
//for the Arduino Environement
//
//=============================================================================
//

#include <Servo.h> 
#include <stdarg.h>

typedef struct ServoMotionStruct {
    int targetAngle;
    float degsPerSecond;
} ServoMotion;

#define MAX_PENDING_MOTIONS 10


typedef struct ServoStateStruct {
    Servo servo;
    int currentAngle;
    int targetAngle;
    int lastMillis;
    float accum;
    float speed;
    // hard coded 10, could change this to whatever a
    // reasonable max value would be, or just use
    // pointers (can't remember whether Arduino can
    // do malloc or not, or whether there are complications
    // for doing so).
    ServoMotion scheduledMotions[MAX_PENDING_MOTIONS];
    int currentMotion;
    int lastMotion;
} ServoState;


#define SERVO_COUNT 9

ServoState servoStates[SERVO_COUNT];

#define EYE_BLINK 0
#define HEAD_ROTATE 1
#define HEAD_UPDOWN 2
#define HEAD_LEFTRIGHT 3
#define NECK_UPDOWN 4
#define NECK_LEFTRIGHT 5
#define WING_RIGHT 6
#define WING_LEFT 7
#define WINGS 8

#define EYE_BLINK_PIN 9
#define HEAD_ROTATE_PIN 5
#define HEAD_UPDOWN_PIN 4
#define HEAD_LEFTRIGHT_PIN 10
#define NECK_UPDOWN_PIN 11
#define NECK_LEFTRIGHT_PIN 8
#define WING_RIGHT_PIN 7
#define WING_LEFT_PIN 6
#define WINGS_PIN 5

int servoPin[] = {
    EYE_BLINK_PIN,
    HEAD_ROTATE_PIN,
    HEAD_UPDOWN_PIN,
    HEAD_LEFTRIGHT_PIN,
    NECK_UPDOWN_PIN,
    NECK_LEFTRIGHT_PIN,
    WING_RIGHT_PIN,
    WING_LEFT_PIN,
    WINGS_PIN
};

int lastMillis = 0;

void addServoMotion(void *statePtr, int angle, float degsPerSecond);
void updateServoPositions();

void p(char *fmt, ... ){
        char tmp[128]; // resulting string limited to 128 chars
        va_list args;
        va_start (args, fmt );
        vsnprintf(tmp, 128, fmt, args);
        va_end (args);
        Serial.print(tmp);
}

int callCount = 0;


void setup() {
    
    Serial.begin(9600); // maybe higher?

    // Initialize the servos, this might be too
    // simple.
    for (int i = 0; i < SERVO_COUNT; i++) {        
        servoStates[i].servo.attach(servoPin[i]);
        servoStates[i].lastMotion = 0;
        servoStates[i].servo.write(5);
        // May need to initialize the other ServoState
        // members here as well!
    }

    // for example
    addServoMotion(&servoStates[0], 150, 60);   // move slowly to 30 degrees
    addServoMotion(&servoStates[0], 30, 60);  // then move more quickly to 150 degrees
    //addServoMotion(&servoStates[1], 30, 1);   // move slowly to 30 degrees
}

void loop() {

    updateServoPositions();
    
}

void addServoMotion(void *statePtr, int angle, float degsPerSecond) {
    ServoState *state = (ServoState *)statePtr;

    state->lastMotion += 1;
    if (state->lastMotion == MAX_PENDING_MOTIONS) {
        // Note: need to check to see whether we're allowed to wrap...
        state->lastMotion = 0;
    }
    
    state->scheduledMotions[state->lastMotion].targetAngle = angle;
    state->scheduledMotions[state->lastMotion].degsPerSecond = degsPerSecond;
}

void updateServoPositions() {
    
    int currentMillis = millis();
    
//    for (int i = 0; i < SERVO_COUNT; i++) {    
    for (int i = 0; i < 1; i++) {    
        ServoState *state = &servoStates[i];
        
        if (state->currentAngle != state->targetAngle) {
            int dir;
            if (state->currentAngle < state->targetAngle) {
                dir = 1;
            } 
            else {
                dir = -1;
            }
            
            Serial.print("  Start");
            Serial.print("(");
            Serial.print(currentMillis);
            Serial.print(")");
            Serial.print(" ");
            Serial.print(i);
            Serial.print(", ");
            Serial.print(state->targetAngle);
            Serial.print(" ");
            Serial.print(state->accum);
            Serial.print(" ");
            
            state->accum += (state->speed * (currentMillis - lastMillis));
            
            Serial.print("new: ");
            Serial.print(state->accum);
            Serial.print(" after ");
            Serial.print(currentMillis - lastMillis);
            Serial.print("ms ");

            Serial.print("(dps = ");
            Serial.print(state->speed);
            Serial.print(") ");
            
            if (state->accum > 1.0) {
                state->currentAngle += dir;
                
                state->servo.write(state->currentAngle);
                
                Serial.print(state->currentAngle);
                Serial.print("' ");
                
                if (state->currentAngle == state->targetAngle) {
                    // We've reached our target angle, throw away any
                    // unused accumulated degrees.
                    state->accum = 0.0;
                    break;
                } else {
                    state->accum -= 1.0;
                }
            }

            Serial.print(state->accum);
            Serial.print(" ");
            Serial.println("End");

        }
        
        // unoptimized
        if (state->currentAngle == state->targetAngle) {
            if (state->currentMotion != state->lastMotion) {
                state->currentMotion += 1;
                if (state->currentMotion > MAX_PENDING_MOTIONS) {
                    state->currentMotion = 0;
                }
                
                state->targetAngle = state->scheduledMotions[state->currentMotion].targetAngle;
                state->speed = state->scheduledMotions[state->currentMotion].degsPerSecond / 1000.0;
        
                p("%d: %d -> %d, %f\r\n", i, state->currentAngle, state->targetAngle, state->scheduledMotions[state->currentMotion].degsPerSecond);
            }
        }    
    }
    
    lastMillis = currentMillis;
}
